"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CrysteamPortal = exports.Methods = void 0;
const axios_1 = __importDefault(require("axios"));
const node_cache_1 = __importDefault(require("node-cache"));
const path_to_regexp_1 = require("path-to-regexp");
const cryptrack_1 = require("@crysteam/cryptrack");
var Methods;
(function (Methods) {
    Methods["POST"] = "POST";
    Methods["PUT"] = "PUT";
    Methods["DELETE"] = "DELETE";
    Methods["PATCH"] = "PATCH";
    Methods["GET"] = "GET";
})(Methods || (exports.Methods = Methods = {}));
class CrysteamPortal {
    constructor(options) {
        this._tokenCache = new node_cache_1.default({ stdTTL: 1800 });
        this._cryptrack = new cryptrack_1.Cryptrack();
        this._hostname = options.hostname;
        this._host = options.host || "https://portal.crysteam.ru/api/v2";
        this._secretKey = Buffer.from(`${options.clientId}:${options.clientSecret}`).toString("base64");
        this._axios = axios_1.default.create({
            baseURL: this._host,
            headers: { authorization: `Bearer ${this._secretKey}` }
        });
    }
    getScopes() {
        return __awaiter(this, void 0, void 0, function* () {
            const token = yield this.getToken();
            const { data } = yield this._axios.get("/services/scopes", {
                headers: { authorization: `Bearer ${token}` }
            });
            return data;
        });
    }
    getToken() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this._ID)
                throw new Error("Missing ID");
            return yield this._cryptrack.signToken(this._ID, {});
        });
    }
    registerService(label_1, description_1) {
        return __awaiter(this, arguments, void 0, function* (label, description, force = false) {
            const { data } = yield this._axios.post("/services/register", {
                label,
                description,
                publicKey: this._cryptrack.getPublicKey(),
                hostname: this._hostname,
            });
            yield this._cryptrack.register(data.id, data.publicKey);
            this._ID = data.id;
        });
    }
    generateAccessToken(serviceFrom, scope) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this._ID)
                throw new Error("Missing ID");
            const cacheKey = `${serviceFrom}:${JSON.stringify(scope)}`;
            const cached = this._tokenCache.get(cacheKey);
            if (cached)
                return cached.token;
            console.log("[Portal] Requesting access token...");
            const { data: token } = yield this._axios.post("/services/produce", {
                service_to: this._ID,
                service_from: serviceFrom,
                scope,
            });
            this._tokenCache.set(cacheKey, { token });
            return token;
        });
    }
    getServiceScopesFor(domain) {
        return __awaiter(this, void 0, void 0, function* () {
            const scopes = yield this.getScopes();
            const domainScopes = scopes[domain];
            if (!domainScopes)
                throw new Error(`[Portal] No scopes found for domain "${domain}"`);
            return domainScopes;
        });
    }
    isScopeAllowed(arg, permission) {
        if (Array.isArray(arg)) {
            return arg.some(scope => scope.method === permission.method &&
                scope.path === permission.path);
        }
        return this.getServiceScopesFor(arg)
            .then(scopes => scopes.some(scope => scope.method === permission.method &&
            scope.path === permission.path))
            .catch(() => false);
    }
    createHttpClientFor(serviceFrom, scope) {
        const client = axios_1.default.create({ baseURL: serviceFrom });
        client.interceptors.request.use((config) => __awaiter(this, void 0, void 0, function* () {
            const token = yield this.generateAccessToken(serviceFrom, scope);
            config.headers['x-service-token'] = `Bearer ${token}`;
            return config;
        }));
        client.interceptors.response.use(undefined, (error) => __awaiter(this, void 0, void 0, function* () {
            var _a;
            if (((_a = error.response) === null || _a === void 0 ? void 0 : _a.status) === 403) {
                const token = yield this.generateAccessToken(serviceFrom, scope);
                error.config.headers['x-service-token'] = `Bearer ${token}`;
                return axios_1.default.request(error.config);
            }
            return Promise.reject(error);
        }));
        return client;
    }
    verifyApiKey() {
        return (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            try {
                if (!this._hostname)
                    throw new Error("Hostname is not configured.");
                const fullPatch = req.baseUrl + req.path;
                console.log(`[Portal][Auth] ${req.method} ${fullPatch}`);
                const serviceToken = req.headers['x-service-token'];
                console.log(serviceToken);
                if (!serviceToken)
                    throw new Error("Missing token.");
                const token = serviceToken.split(" ")[1];
                if (!token)
                    throw new Error("Missing token.");
                const { payload } = yield this._cryptrack.verifyToken(token);
                console.log(payload);
                const servicePayload = payload;
                const isAllowed = servicePayload.scope.some(({ method, path }) => {
                    if (servicePayload.service_from !== this._hostname)
                        return false;
                    if (method !== req.method)
                        return false;
                    return (0, path_to_regexp_1.match)(path, { decode: decodeURIComponent })(fullPatch) !== false;
                });
                if (!isAllowed)
                    throw new Error("Access denied.");
                req.portal = servicePayload;
                next();
            }
            catch (err) {
                next(err);
            }
        });
    }
}
exports.CrysteamPortal = CrysteamPortal;
