import axios, { AxiosInstance } from "axios";
import NodeCache from "node-cache";
import { match } from "path-to-regexp";
import { RequestHandler } from "express";
import { Cryptrack } from "@crysteam/cryptrack";

export enum Methods { POST = 'POST', PUT = 'PUT', DELETE = 'DELETE', PATCH = 'PATCH', GET = 'GET' }

export type RoutePermission = { method: Methods; path: string };
export type Scopes = { [domain: string]: RoutePermission[] };
export type JwtPayload = { service_from: string; service_to: string; scope: RoutePermission[] };
export type ServiceMeta = { clientPrivateKey: string; portalPublicKey: string; serviceId: string };

export type PortalOptions = {
    hostname?: string;
    clientId: string;
    clientSecret: string;
    label?: string;
    description?: string;
    host?: string
};

export class CrysteamPortal {
    private _ID: string | undefined;
    private readonly _tokenCache = new NodeCache({ stdTTL: 1800 });
    private readonly _cryptrack: Cryptrack = new Cryptrack();

    private readonly _axios: AxiosInstance;
    private readonly _hostname?: string;
    private readonly _host: string;
    private readonly _secretKey: string;

    constructor(options: PortalOptions) {
        this._hostname = options.hostname;
        this._host = options.host || "https://portal.crysteam.ru/api/v2";

        this._secretKey = Buffer.from(`${options.clientId}:${options.clientSecret}`).toString("base64");
        this._axios = axios.create({
            baseURL: this._host,
            headers: { authorization: `Bearer ${this._secretKey}` }
        });
    }

    async getScopes(): Promise<Scopes> {
        const token = await this.getToken();
        const { data } = await this._axios.get("/services/scopes", {
            headers: { authorization: `Bearer ${token}` }
        });
        return data;
    }

    async getToken(): Promise<string> {
        if (!this._ID) throw new Error("Missing ID");
        return await this._cryptrack.signToken(this._ID, {});
    }

    async registerService(label: string, description?: string, force: boolean = false): Promise<void> {
        const { data } = await this._axios.post("/services/register", {
            label,
            description,
            publicKey: this._cryptrack.getPublicKey(),
            hostname: this._hostname,
        });

        await this._cryptrack.register(data.id, data.publicKey);
        this._ID = data.id;
    }

    async generateAccessToken(serviceFrom: string, scope: RoutePermission[]): Promise<string> {
        if (!this._ID) throw new Error("Missing ID");

        const cacheKey = `${serviceFrom}:${JSON.stringify(scope)}`;
        const cached = this._tokenCache.get<{ token: string }>(cacheKey);
        if (cached) return cached.token;

        console.log("[Portal] Requesting access token...");

        const { data: token } = await this._axios.post("/services/produce", {
            service_to: this._ID,
            service_from: serviceFrom,
            scope,
        });

        this._tokenCache.set(cacheKey, { token });
        return token;
    }

    async getServiceScopesFor(domain: string): Promise<RoutePermission[]> {
        const scopes = await this.getScopes();
        const domainScopes = scopes[domain];
        if (!domainScopes) throw new Error(`[Portal] No scopes found for domain "${domain}"`);
        return domainScopes;
    }

    isScopeAllowed(scopes: RoutePermission[], permission: RoutePermission): boolean;
    isScopeAllowed(domain: string, permission: RoutePermission): Promise<boolean>;
    isScopeAllowed(arg: string | RoutePermission[], permission: RoutePermission): boolean | Promise<boolean> {
        if (Array.isArray(arg)) {
            return arg.some(scope =>
                scope.method === permission.method &&
                scope.path === permission.path
            );
        }

        return this.getServiceScopesFor(arg)
            .then(scopes =>
                scopes.some(scope =>
                    scope.method === permission.method &&
                    scope.path === permission.path
                )
            )
            .catch(() => false);
    }

    createHttpClientFor(serviceFrom: string, scope: RoutePermission[]) {
        const client = axios.create({ baseURL: serviceFrom });

        client.interceptors.request.use(async config => {
            const token = await this.generateAccessToken(serviceFrom, scope);
            config.headers['x-service-token'] = `Bearer ${token}`;
            return config;
        });

        client.interceptors.response.use(undefined, async error => {
            if (error.response?.status === 403) {
                const token = await this.generateAccessToken(serviceFrom, scope);
                error.config.headers['x-service-token'] = `Bearer ${token}`;
                return axios.request(error.config);
            }
            return Promise.reject(error);
        });

        return client;
    }

    verifyApiKey(): RequestHandler {
        return async (req, res, next) => {
            try {
                if (!this._hostname) throw new Error("Hostname is not configured.");
                const fullPatch = req.baseUrl + req.path;
                console.log(`[Portal][Auth] ${req.method} ${fullPatch}`);

                const serviceToken = req.headers['x-service-token'] as string | undefined;
                console.log(serviceToken)
                if (!serviceToken) throw new Error("Missing token.");

                const token = serviceToken.split(" ")[1];
                if (!token) throw new Error("Missing token.");

                const { payload } = await this._cryptrack.verifyToken(token);
                console.log(payload)
                const servicePayload = payload as JwtPayload;

                const isAllowed = servicePayload.scope.some(({ method, path }) => {
                    if (servicePayload.service_from !== this._hostname) return false;
                    if (method !== req.method) return false;
                    return match(path, { decode: decodeURIComponent })(fullPatch) !== false;
                });

                if (!isAllowed) throw new Error("Access denied.");
                (req as any).portal = servicePayload;
                next();
            } catch (err) {
                next(err);
            }
        };
    }
}
